import asyncio
import pytest

from kafka.errors import KafkaError
from kafkahelpers import ReconnectingClient


class FakeMQ:
    async def start(self):
        return True


class FakeMQFails:
    failures = 0

    async def start(self):
        if not self.failures:
            self.failures = 1
            raise KafkaError("fakemq fails")

        else:
            return True


@pytest.mark.asyncio
async def test_connects():
    mq = FakeMQ()
    cl = ReconnectingClient(mq, "connects")
    await cl.start()
    assert cl.connected == True


@pytest.mark.asyncio
async def test_reconnects():
    mq = FakeMQFails()
    cl = ReconnectingClient(mq, "reconnects", retry_interval=0.1)
    await cl.start()
    assert cl.connected == True


async def wrkr(client):
    return True


async def wrkr_fails(client):
    print(client.failures)
    if client.failures < 2:
        client.failures += 1
        raise KafkaError("wrkr_fails")

    return True


@pytest.mark.asyncio
async def test_work():
    mq = FakeMQ()
    cl = ReconnectingClient(mq, "works")
    await cl.work(wrkr)


@pytest.mark.asyncio
async def test_work_fails():
    mq = FakeMQFails()
    cl = ReconnectingClient(mq, "works_fails", retry_interval=0.1)
    await cl.work(wrkr_fails)
    await cl.start()
    assert cl.connected == True


def done(v):
    return v == False


@pytest.mark.asyncio
async def test_callback(event_loop):
    mq = FakeMQ()
    cl = ReconnectingClient(mq, "callback")
    coro = cl.get_callback(wrkr, done)
    task = event_loop.create_task(coro())
    await task
    assert task.result() == None


@pytest.mark.asyncio
async def test_callback_retries(event_loop):
    mq = FakeMQFails()
    cl = ReconnectingClient(mq, "callback_retries", retry_interval=0.1)
    coro = cl.get_callback(wrkr_fails, done)
    task = event_loop.create_task(coro())
    await task
    assert task.result() == None


@pytest.mark.asyncio
async def test_run(event_loop):
    mq = FakeMQ()
    cl = ReconnectingClient(mq, "callback")
    task = event_loop.create_task(cl.run(wrkr, done))
    await task
    assert task.result() == None



import asyncio
import logging
import attr
from kafka.errors import KafkaError

loop = asyncio.get_event_loop()
logger = logging.getLogger(__name__)


@attr.s
class ReconnectingClient(object):
    """
    Wraps an aiokafka client such that it will reconnect forever.

    Provides a way to add worker functions to use the wrapped client forever.

    example:

        client = ReconnectingClient(aiokafka.AIOKafkaConsumer(), "consumer")
        loop.create_task(client.run(my_consumer_func))
    """

    client = attr.ib()
    name = attr.ib()
    retry_interval = attr.ib(default=5)
    connected = attr.ib(init=False, default=False)

    async def start(self):
        """
        Connects the wrapped client.
        """
        while not self.connected:
            try:
                logger.info("Attempting to connect %s client.", self.name)
                await self.client.start()
                logger.info("%s client connected successfully.", self.name)
                self.connected = True
            except KafkaError:
                logger.exception(
                    "Failed to connect %s client, retrying in %d seconds.",
                    self.name,
                    self.retry_interval,
                )
                await asyncio.sleep(self.retry_interval)

    async def work(self, worker):
        """
        Executes the worker function.
        """
        try:
            await worker(self.client)
        except KafkaError:
            logger.exception(
                "Encountered exception while working %s client, reconnecting.",
                self.name,
            )
            self.connected = False

    def get_callback(self, worker, cond=lambda v: True):
        """
        Returns a callback function that will ensure that the wrapped client is
        connected forever.

        example:

            loop.spawn_callback(client.get_callback(my_cb))
        """

        async def _f():
            v = cond(None)
            while cond(v):
                await self.start()
                v = await self.work(worker)

        return _f

    def run(self, worker, cond=lambda v: True):
        """
        Returns a coroutine that will ensure that the wrapped client is
        connected forever.

        example:

            loop.create_task(client.run(my_func))
        """
        return self.get_callback(worker, cond)()


def consumer(queue, bootstrap_servers, group_id, name="reader", loop=loop):
    """
    consumer returns a wrapped kafka consumer that will reconnect.
    """
    return ReconnectingClient(
        aiokafka.AIOKafkaConsumer(
            queue, bootstrap_servers=bootstrap_servers, group_id=group_id, loop=loop
        ),
        name,
    )


def producer(
    bootstrap_servers,
    loop=loop,
    request_timeout_ms=10000,
    connections_max_idle_ms=None,
    name="writer",
):
    """
    producer returns a wrapped kafka producer that will reconnect.
    """
    return ReconnectingClient(
        aiokafka.AIOKafkaProducer(
            loop=loop,
            request_timeout_ms=request_timeout_ms,
            connections_max_idle_ms=connections_max_idle_ms,
        ),
        name,
    )


def make_pair(
    read_queue,
    read_group_id,
    bootstrap_servers,
    loop=loop,
    request_timeout_ms=10000,
    connections_max_idle_ms=None,
    read_name="reader",
    write_name="writer",
):
    """
    make_pair returns a consumer, producer tuple that will reconnect.
    """
    return (
        consumer(
            read_queue, bootstrap_servers, read_group_id, name=read_name, loop=loop
        ),
        producer(
            bootstrap_servers,
            loop=loop,
            request_timeout_ms=request_timeout_ms,
            connections_max_idle_ms=connections_max_idle_ms,
        ),
    )



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['kafkahelpers']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=18.2,<19.0', 'kafka-python>=1.4,<2.0']

setup_kwargs = {
    'name': 'kafkahelpers',
    'version': '0.2.0',
    'description': '',
    'long_description': None,
    'author': 'Jesse Jaggars',
    'author_email': 'jhjaggars@gmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
